#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_shadow.h>
#include <msp/gl/extensions/arb_texture_swizzle.h>
#include <msp/gl/extensions/ext_framebuffer_object.h>
#include <msp/gl/extensions/ext_texture3d.h>
#include <msp/gl/extensions/ext_texture_filter_anisotropic.h>
#include <msp/gl/extensions/sgis_generate_mipmap.h>
#include <msp/io/memory.h>
#include <msp/strings/format.h>
#include "error.h"
#include "resourcemanager.h"
#include "resources.h"
#include "texture.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

void operator>>(const LexicalConverter &c, TextureFilter &tf)
{
	if(c.get()=="NEAREST")
		tf = NEAREST;
	else if(c.get()=="LINEAR")
		tf = LINEAR;
	else if(c.get()=="NEAREST_MIPMAP_NEAREST")
		tf = NEAREST_MIPMAP_NEAREST;
	else if(c.get()=="NEAREST_MIPMAP_LINEAR")
		tf = NEAREST_MIPMAP_LINEAR;
	else if(c.get()=="LINEAR_MIPMAP_NEAREST")
		tf = LINEAR_MIPMAP_NEAREST;
	else if(c.get()=="LINEAR_MIPMAP_LINEAR")
		tf = LINEAR_MIPMAP_LINEAR;
	else
		throw lexical_error(format("conversion of '%s' to TextureFilter", c.get()));
}


void operator>>(const LexicalConverter &c, TextureWrap &tw)
{
	if(c.get()=="REPEAT")
		tw = REPEAT;
	else if(c.get()=="CLAMP_TO_EDGE")
		tw = CLAMP_TO_EDGE;
	else if(c.get()=="MIRRORED_REPEAT")
		tw = MIRRORED_REPEAT;
	else
		throw lexical_error(format("conversion of '%s' to TextureWrap", c.get()));
}


int Texture::swizzle_orders[] =
{
	GL_RED, GL_GREEN, GL_BLUE, GL_ALPHA,
	GL_RED, GL_RED, GL_RED, GL_ONE,
	GL_RED, GL_RED, GL_RED, GL_GREEN
};

Texture::Texture(GLenum t, ResourceManager *m):
	id(0),
	target(t),
	ifmt(RGB),
	min_filter(NEAREST_MIPMAP_LINEAR),
	mag_filter(LINEAR),
	wrap_s(REPEAT),
	wrap_t(REPEAT),
	wrap_r(REPEAT),
	gen_mipmap(false),
	compare(false),
	cmp_func(LEQUAL),
	dirty_params(0)
{
	if(m)
		set_manager(m);
	else if(ARB_direct_state_access)
		glCreateTextures(target, 1, &id);
	else
		glGenTextures(1, &id);
}

Texture::~Texture()
{
	while(TexUnit *unit = TexUnit::find_unit(this))
		unbind_from(unit->get_index());

	if(id)
		glDeleteTextures(1, &id);
}

DataType Texture::get_alloc_type(PixelFormat fmt)
{
	return (get_base_pixelformat(fmt)==DEPTH_COMPONENT ? UNSIGNED_SHORT : UNSIGNED_BYTE);
}

void Texture::set_internal_format(PixelFormat fmt)
{
	if(!get_component_size(fmt) && MSP_sized_internal_formats)
	{
		unsigned size = (fmt==DEPTH_COMPONENT ? get_gl_api()==OPENGL_ES2 ? 2 : 4 : 1);
		fmt = get_sized_pixelformat(fmt, size);
	}

	FormatSwizzle swiz = NO_SWIZZLE;
	if(ARB_texture_rg && ARB_texture_swizzle)
	{
		if(fmt==LUMINANCE8 || fmt==SLUMINANCE8)
		{
			fmt = R8;
			swiz = R_TO_LUMINANCE;
		}
		else if(fmt==LUMINANCE8_ALPHA8 || fmt==SLUMINANCE8_ALPHA8)
		{
			fmt = RG8;
			swiz = RG_TO_LUMINANCE_ALPHA;
		}
	}

	require_pixelformat(fmt);
	ifmt = fmt;
	swizzle = swiz;
	if(swizzle)
		update_parameter(FORMAT_SWIZZLE);
}

PixelFormat Texture::get_upload_format(PixelFormat fmt) const
{
	if(fmt==LUMINANCE || fmt==LUMINANCE_ALPHA)
		return get_base_pixelformat(ifmt);
	else
		return fmt;
}

void Texture::update_parameter(int mask) const
{
	if(!ARB_direct_state_access && TexUnit::current().get_texture()!=this)
	{
		TexUnit *unit = TexUnit::find_unit(this);
		if(!unit)
		{
			dirty_params |= mask;
			return;
		}

		unit->bind();
	}

	if(mask&MIN_FILTER)
		set_parameter_i(GL_TEXTURE_MIN_FILTER, min_filter);
	if(mask&MAG_FILTER)
		set_parameter_i(GL_TEXTURE_MAG_FILTER, mag_filter);
	if(mask&MAX_ANISOTROPY)
		set_parameter_f(GL_TEXTURE_MAX_ANISOTROPY_EXT, max_anisotropy);
	if(mask&WRAP_S)
		set_parameter_i(GL_TEXTURE_WRAP_S, wrap_s);
	if(mask&WRAP_T)
		set_parameter_i(GL_TEXTURE_WRAP_T, wrap_t);
	if(mask&WRAP_R)
		set_parameter_i(GL_TEXTURE_WRAP_R, wrap_r);
	if(mask&GENERATE_MIPMAP)
		set_parameter_i(GL_GENERATE_MIPMAP, gen_mipmap);
	if(mask&COMPARE)
		set_parameter_i(GL_TEXTURE_COMPARE_MODE, (compare ? GL_COMPARE_R_TO_TEXTURE : GL_NONE));
	if(mask&COMPARE_FUNC)
		set_parameter_i(GL_TEXTURE_COMPARE_FUNC, cmp_func);
	if(mask&FORMAT_SWIZZLE)
	{
		if(ARB_direct_state_access)
			glTextureParameteriv(id, GL_TEXTURE_SWIZZLE_RGBA, swizzle_orders+swizzle*4);
		else
			glTexParameteriv(target, GL_TEXTURE_SWIZZLE_RGBA, swizzle_orders+swizzle*4);
	}
}

void Texture::set_parameter_i(GLenum param, int value) const
{
	if(ARB_direct_state_access)
		glTextureParameteri(id, param, value);
	else
		glTexParameteri(target, param, value);
}

void Texture::set_parameter_f(GLenum param, float value) const
{
	if(ARB_direct_state_access)
		glTextureParameterf(id, param, value);
	else
		glTexParameterf(target, param, value);
}

void Texture::set_min_filter(TextureFilter f)
{
	min_filter = f;
	update_parameter(MIN_FILTER);
}

void Texture::set_mag_filter(TextureFilter f)
{
	mag_filter = f;
	update_parameter(MAG_FILTER);
}

void Texture::set_filter(TextureFilter f)
{
	set_min_filter(f);
	set_mag_filter(f==NEAREST ? NEAREST : LINEAR);
}

void Texture::set_max_anisotropy(float a)
{
	if(a<1.0f)
		throw invalid_argument("Texture::set_max_anisotropy");
	else if(a>1.0f)
		static Require _req(EXT_texture_filter_anisotropic);
	max_anisotropy = a;
	update_parameter(MAX_ANISOTROPY);
}

void Texture::set_wrap(TextureWrap w)
{
	set_wrap_s(w);
	set_wrap_t(w);
	if(EXT_texture3D)
		set_wrap_r(w);
}

void Texture::set_wrap_s(TextureWrap w)
{
	wrap_s = w;
	update_parameter(WRAP_S);
}

void Texture::set_wrap_t(TextureWrap w)
{
	wrap_t = w;
	update_parameter(WRAP_T);
}

void Texture::set_wrap_r(TextureWrap w)
{
	static Require _req(EXT_texture3D);
	wrap_r = w;
	update_parameter(WRAP_R);
}

bool Texture::can_generate_mipmap()
{
	return (EXT_framebuffer_object || SGIS_generate_mipmap);
}

void Texture::set_generate_mipmap(bool gm)
{
	if(gm && !EXT_framebuffer_object)
		static Require _req(SGIS_generate_mipmap);
	gen_mipmap = gm;
	if(!EXT_framebuffer_object)
		update_parameter(GENERATE_MIPMAP);
}

void Texture::auto_generate_mipmap()
{
	// glGenerateMipmap is defined here
	if(EXT_framebuffer_object)
	{
		if(ARB_direct_state_access)
			glGenerateTextureMipmap(id);
		else
			glGenerateMipmap(target);
	}
}

void Texture::set_compare_enabled(bool c)
{
	if(c)
		static Require _req(ARB_shadow);
	compare = c;
	update_parameter(COMPARE);
}

void Texture::set_compare_func(Predicate f)
{
	static Require _req(ARB_shadow);
	cmp_func = f;
	update_parameter(COMPARE_FUNC);
}

void Texture::load_image(const string &fn, bool srgb)
{
	Graphics::Image img;
	img.load_file(fn);

	image(img, srgb);
}

void Texture::bind_to(unsigned i) const
{
	if(!id)
	{
		if(manager)
			manager->resource_used(*this);
		if(!id)
		{
			unbind_from(i);
			return;
		}
	}

	TexUnit &unit = TexUnit::get_unit(i);
	const Texture *old = unit.get_texture();
	if(unit.set_texture(this))
	{
		if(manager)
			manager->resource_used(*this);

		unit.bind();
		if(unit.supports_legacy())
		{
			if(old && old->target!=target)
				glDisable(old->target);
			if(!old || old->target!=target)
				glEnable(target);
		}
		glBindTexture(target, id);

		if(dirty_params)
		{
			update_parameter(dirty_params);
			dirty_params = 0;
		}
	}
}

const Texture *Texture::current(unsigned i)
{
	return TexUnit::get_unit(i).get_texture();
}

void Texture::unbind_from(unsigned i)
{
	TexUnit &unit = TexUnit::get_unit(i);
	const Texture *cur = unit.get_texture();
	if(unit.set_texture(0))
	{
		unit.bind();
		glBindTexture(cur->target, 0);
		if(unit.supports_legacy())
			glDisable(cur->target);
	}
}


Texture::Loader::Loader(Texture &t):
	DataFile::CollectionObjectLoader<Texture>(t, 0)
{
	init();
}

Texture::Loader::Loader(Texture &t, Collection &c):
	DataFile::CollectionObjectLoader<Texture>(t, &c)
{
	init();
}

void Texture::Loader::init()
{
	if(Resources *res = dynamic_cast<Resources *>(coll))
		srgb = res->get_srgb_conversion();
	else
		srgb = false;

	add("external_image", &Loader::external_image);
	add("filter", &Loader::filter);
	add("generate_mipmap", &Loader::generate_mipmap);
	add("image_data", &Loader::image_data);
	add("mag_filter", &Loader::mag_filter);
	add("max_anisotropy", &Loader::max_anisotropy);
	add("min_filter", &Loader::min_filter);
	add("wrap",       &Loader::wrap);
	add("wrap_r",     &Loader::wrap_r);
	add("wrap_s",     &Loader::wrap_s);
	add("wrap_t",     &Loader::wrap_t);
}

void Texture::Loader::external_image(const string &fn)
{
	Graphics::Image img;
	RefPtr<IO::Seekable> io = get_collection().open_raw(fn);
	img.load_io(*io);

	obj.image(img, srgb);
}

void Texture::Loader::filter(TextureFilter f)
{
	obj.set_filter(f);
}

void Texture::Loader::generate_mipmap(bool gm)
{
	obj.set_generate_mipmap(gm);
}

void Texture::Loader::image_data(const string &data)
{
	Graphics::Image img;
	IO::Memory mem(data.data(), data.size());
	img.load_io(mem);

	obj.image(img, srgb);
}

void Texture::Loader::mag_filter(TextureFilter f)
{
	obj.set_mag_filter(f);
}

void Texture::Loader::max_anisotropy(float a)
{
	obj.set_max_anisotropy(a);
}

void Texture::Loader::min_filter(TextureFilter f)
{
	obj.set_min_filter(f);
}

void Texture::Loader::wrap(TextureWrap w)
{
	obj.set_wrap(w);
}

void Texture::Loader::wrap_r(TextureWrap w)
{
	obj.set_wrap_r(w);
}

void Texture::Loader::wrap_s(TextureWrap w)
{
	obj.set_wrap_s(w);
}

void Texture::Loader::wrap_t(TextureWrap w)
{
	obj.set_wrap_t(w);
}


bool is_mipmapped(TextureFilter filter)
{
	return (filter==NEAREST_MIPMAP_NEAREST || filter==NEAREST_MIPMAP_LINEAR ||
		filter==LINEAR_MIPMAP_NEAREST || filter==LINEAR_MIPMAP_LINEAR);
}

} // namespace GL
} // namespace Msp
