#ifndef MSP_GL_TEXTURE2D_H_
#define MSP_GL_TEXTURE2D_H_

#include <string>
#include <msp/graphics/image.h>
#include "texture.h"

namespace Msp {
namespace GL {

/**
Two-dimensional texture.  Consists of an array of texels in the shape of a
rectangle.  Texture coordinate have a range of [0, 1].  Coordinates outside of
this range are subject to wrapping.  This is the most common type of texture.
*/
class Texture2D: public Texture
{
public:
	class Loader: public Msp::DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>
	{
	public:
		Loader(Texture2D &);
		Loader(Texture2D &, Collection &);
	private:
		void init();

		void raw_data(const std::string &);
		void storage(PixelFormat, unsigned, unsigned);
	};

private:
	class AsyncLoader;

	unsigned width;
	unsigned height;
	unsigned allocated;

public:
	Texture2D(ResourceManager * = 0);
	virtual ~Texture2D();

	/** Defines storage structure for the texture.  Must be called before an
	image can be uploaded.  Once storage is defined, it can't be changed. */
	void storage(PixelFormat fmt, unsigned wd, unsigned ht);

	/** Allocates storage for the texture.  The contents are initially
	undefined.  If storage has already been allocated, does nothing. */
	void allocate(unsigned level);

	/** Uploads an image to the texture.  Storage must be defined beforehand.
	The image data must have dimensions and format compatible with the defined
	storage. */
	void image(unsigned level, PixelFormat fmt, DataType type, const void *data);

	/** Updates a rectangular region of the texture.  Storage must be defined
	and allocated beforehand.  The update region must be fully inside the
	texture. */
	void sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht,
		PixelFormat fmt, DataType type, const void *data);

	/** Uploads an image to the texture.  If storage has not been defined, it
	will be set to match the image.  Otherwise the image must be compatible with
	the defined storage.

	If srgb is true and storage is determined by this call, then an sRGB pixel
	format will be used. */
	virtual void image(const Graphics::Image &, bool srgb = false);

private:
	void image(const Graphics::Image &, bool, bool);

public:
	unsigned get_width() const { return width; }
	unsigned get_height() const { return height; }

private:
	unsigned get_n_levels() const;
	void get_level_size(unsigned, unsigned &, unsigned &) const;

public:
	virtual Resource::AsyncLoader *load(IO::Seekable &, const Resources * = 0);
	virtual UInt64 get_data_size() const;
	virtual void unload();
};

} // namespace GL
} // namespace Msp

#endif
